import pytest

import pygeoutils as utils
from pygeoutils import InvalidInputType, InvalidInputValue

DEF_CRS = "epsg:4326"


def test_invalid_json2geodf_type():
    with pytest.raises(InvalidInputType) as ex:
        _ = utils.json2geodf("")
    assert "content" in str(ex.value)


def test_json2geodf_empty():
    with pytest.raises(StopIteration) as ex:
        _ = utils.json2geodf([])
    assert "Resnpose is empty." in str(ex.value)


def test_gtiff2xarray_type():
    with pytest.raises(InvalidInputType) as ex:
        _ = utils.gtiff2xarray([], (0, 0, 0, 0), DEF_CRS)
    assert "Response.content" in str(ex.value)


def test_gtiff2xarray_empty():
    with pytest.raises(StopIteration) as ex:
        _ = utils.gtiff2xarray({}, (0, 0, 0, 0), DEF_CRS)
    assert "Resnpose dict is empty." in str(ex.value)



"""Top-level package for PyGeoUtils."""
from pkg_resources import DistributionNotFound, get_distribution

from .exceptions import InvalidInputType, InvalidInputValue
from .print_versions import show_versions
from .pygeoutils import MatchCRS, arcgis2geojson, gtiff2xarray, json2geodf

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    # package is not installed
    pass



"""Customized PyGeoUtils exceptions."""
from typing import Generator, List, Optional, Union


class InvalidInputType(Exception):
    """Exception raised when a function argument type is invalid.

    Parameters
    ----------
    arg : str
        Name of the function argument
    valid_type : str
        The valid type of the argument
    example : str, optional
        An example of a valid form of the argument, defaults to None.
    """

    def __init__(self, arg: str, valid_type: str, example: Optional[str] = None) -> None:
        self.message = f"The {arg} argument should be of type {valid_type}"
        if example is not None:
            self.message += f":\n{example}"
        super().__init__(self.message)

    def __str__(self) -> str:
        return self.message


class InvalidInputValue(Exception):
    """Exception raised for invalid input.

    Parameters
    ----------
    inp : str
        Name of the input parameter
    valid_inputs : tuple
        List of valid inputs
    """

    def __init__(
        self, inp: str, valid_inputs: Union[List[str], Generator[str, None, None]]
    ) -> None:
        self.message = f"Given {inp} is invalid. Valid {inp}s are:\n" + ", ".join(
            str(i) for i in valid_inputs
        )
        super().__init__(self.message)

    def __str__(self) -> str:
        return self.message



"""Some utilities for manipulating GeoSpatial data."""
import contextlib
import logging
import numbers
import sys
import tempfile
import uuid
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

import affine
import geopandas as gpd
import numpy as np
import orjson as json
import pyproj
import rasterio as rio
import rasterio.mask as rio_mask
import rasterio.transform as rio_transform
import shapely.geometry as sgeom
import xarray as xr
from shapely import ops
from shapely.geometry import LineString, MultiLineString, MultiPoint, MultiPolygon, Point, Polygon

from .exceptions import InvalidInputType, InvalidInputValue

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
handler = logging.StreamHandler(sys.stdout)
handler.setFormatter(logging.Formatter(""))
logger.handlers = [handler]
logger.propagate = False

DEF_CRS = "epsg:4326"
BOX_ORD = "(west, south, east, north)"


def json2geodf(
    content: Union[List[Dict[str, Any]], Dict[str, Any]],
    in_crs: str = DEF_CRS,
    crs: str = DEF_CRS,
) -> gpd.GeoDataFrame:
    """Create GeoDataFrame from (Geo)JSON.

    Parameters
    ----------
    content : dict or list of dict
        A (Geo)JSON dictionary e.g., r.json() or a list of them.
    in_crs : str
        CRS of the content, defaults to ``epsg:4326``.
    crs : str, optional
        The target CRS of the output GeoDataFrame, defaults to ``epsg:4326``.

    Returns
    -------
    geopandas.GeoDataFrame
        Generated geo-data frame from a GeoJSON
    """
    if not isinstance(content, (list, dict)):
        raise InvalidInputType("content", "list or list of dict ((geo)json)")

    content = content if isinstance(content, list) else [content]
    try:
        geodf = gpd.GeoDataFrame.from_features(next(iter(content)))
    except TypeError:
        content = [arcgis2geojson(c) for c in content]
        geodf = gpd.GeoDataFrame.from_features(content[0])
    except StopIteration:
        raise StopIteration("Resnpose is empty.")

    if len(content) > 1:
        geodf = geodf.append([gpd.GeoDataFrame.from_features(c) for c in content[1:]])

    if "geometry" in geodf and len(geodf) > 0:
        geodf = geodf.set_crs(in_crs)
        if in_crs != crs:
            geodf = geodf.to_crs(crs)

    return geodf


def arcgis2geojson(
    arcgis: Union[str, Dict[str, Any]], id_attr: Optional[str] = None
) -> Dict[str, Any]:
    """Convert ESRIGeoJSON format to GeoJSON.

    Notes
    -----
    Based on `arcgis2geojson <https://github.com/chris48s/arcgis2geojson>`__.

    Parameters
    ----------
    arcgis : str or binary
        The ESRIGeoJSON format str (or binary)
    id_attr : str
        ID of the attribute of interest

    Returns
    -------
    dict
        A GeoJSON file readable by GeoPandas.
    """

    def convert(arcgis: Dict[str, Any], id_attr: Optional[str] = None) -> Dict[str, Any]:
        """Convert an ArcGIS JSON object to a GeoJSON object."""
        geojson: Dict[str, Any] = {}

        if arcgis.get("features") is not None:
            geojson["type"] = "FeatureCollection"
            geojson["features"] = [convert(f, id_attr) for f in arcgis["features"]]

        if isinstance(arcgis.get("x"), numbers.Number) and isinstance(
            arcgis.get("y"), numbers.Number
        ):
            geojson["type"] = "Point"
            geojson["coordinates"] = [arcgis["x"], arcgis["y"]]
            if isinstance(arcgis.get("z"), numbers.Number):
                geojson["coordinates"].append(arcgis["z"])

        if arcgis.get("points") is not None:
            geojson["type"] = "MultiPoint"
            geojson["coordinates"] = arcgis["points"]

        if arcgis.get("paths") is not None:
            if len(arcgis["paths"]) == 1:
                geojson["type"] = "LineString"
                geojson["coordinates"] = arcgis["paths"][0]
            else:
                geojson["type"] = "MultiLineString"
                geojson["coordinates"] = arcgis["paths"]

        if arcgis.get("rings") is not None:
            geojson = _rings2geojson(arcgis["rings"])

        coords = ("xmin", "xmax", "ymin", "ymax")
        if all(isinstance(arcgis.get(c), numbers.Number) for c in coords):
            geojson["type"] = "Polygon"
            geojson["coordinates"] = [
                [
                    [arcgis["xmax"], arcgis["ymax"]],
                    [arcgis["xmin"], arcgis["ymax"]],
                    [arcgis["xmin"], arcgis["ymin"]],
                    [arcgis["xmax"], arcgis["ymin"]],
                    [arcgis["xmax"], arcgis["ymax"]],
                ]
            ]

        if arcgis.get("geometry") is not None or arcgis.get("attributes") is not None:
            geojson["type"] = "Feature"
            if arcgis.get("geometry") is not None:
                geojson["geometry"] = convert(arcgis["geometry"])
            else:
                geojson["geometry"] = None

            if arcgis.get("attributes") is not None:
                geojson["properties"] = arcgis["attributes"]
                keys = [id_attr, "OBJECTID", "FID"] if id_attr else ["OBJECTID", "FID"]
                for key in keys:
                    if isinstance(arcgis["attributes"].get(key), (numbers.Number, str)):
                        geojson["id"] = arcgis["attributes"][key]
                        break
                if "id" not in geojson:
                    logger.warning("No valid id attribute found.")
            else:
                geojson["properties"] = None

        if json.dumps(geojson.get("geometry")) == json.dumps({}):
            geojson["geometry"] = None

        return geojson

    if isinstance(arcgis, str):
        return convert(json.loads(arcgis), id_attr)

    return convert(arcgis, id_attr)


def _rings2geojson(rings: List[List[List[float]]]) -> Dict[str, Any]:
    """Check for holes in the ring and fill them."""
    outer_rings, holes = _get_outer_rings(rings)
    uncontained_holes = _get_uncontained_holes(outer_rings, holes)

    # if we couldn't match any holes using contains we can try intersects...
    while uncontained_holes:
        # pop a hole off out stack
        hole = uncontained_holes.pop()

        # loop over all outer rings and see if any intersect our hole.
        intersects = False
        x = len(outer_rings) - 1
        while x >= 0:
            outer_ring = outer_rings[x][0]
            l1, l2 = LineString(outer_ring), LineString(hole)
            intersects = l1.intersects(l2)
            if intersects:
                # the hole is contained push it into our polygon
                outer_rings[x].append(hole)  # type: ignore
                intersects = True
                break
            x = x - 1

        if not intersects:
            outer_rings.append([hole[::-1]])  # type: ignore

    if len(outer_rings) == 1:
        return {"type": "Polygon", "coordinates": outer_rings[0]}

    return {"type": "MultiPolygon", "coordinates": outer_rings}


def _get_outer_rings(rings: List[List[List[float]]]) -> Tuple[List[List[float]], List[List[float]]]:
    """Get outer rings and holes in a list of rings."""
    outer_rings = []
    holes = []
    for ring in rings:
        if not np.all(np.isclose(ring[0], ring[-1])):
            ring.append(ring[0])

        if len(ring) < 4:
            continue

        total = sum((pt2[0] - pt1[0]) * (pt2[1] + pt1[1]) for pt1, pt2 in zip(ring[:-1], ring[1:]))
        # Clock-wise check
        if total >= 0:
            # wind outer rings counterclockwise for RFC 7946 compliance
            outer_rings.append([ring[::-1]])
        else:
            # wind inner rings clockwise for RFC 7946 compliance
            holes.append(ring[::-1])
    return outer_rings, holes  # type: ignore


def _get_uncontained_holes(
    outer_rings: List[List[float]], holes: List[List[float]]
) -> List[List[float]]:
    """Get all the uncontstrained holes."""
    uncontained_holes = []

    # while there are holes left...
    while holes:
        # pop a hole off out stack
        hole = holes.pop()

        # loop over all outer rings and see if they contain our hole.
        contained = False
        x = len(outer_rings) - 1
        while x >= 0:
            outer_ring = outer_rings[x][0]
            l1, l2 = LineString(outer_ring), LineString(hole)
            p2 = Point(hole[0])
            intersects = l1.intersects(l2)
            contains = l1.contains(p2)
            if not intersects and contains:
                # the hole is contained push it into our polygon
                outer_rings[x].append(hole)  # type: ignore
                contained = True
                break
            x = x - 1

        # ring is not contained in any outer ring
        # sometimes this happens https://github.com/Esri/esri-leaflet/issues/320
        if not contained:
            uncontained_holes.append(hole)
    return uncontained_holes


def gtiff2xarray(
    r_dict: Dict[str, bytes],
    geometry: Union[Polygon, MultiPolygon, Tuple[float, float, float, float]],
    geo_crs: str,
    ds_dims: Tuple[str, str] = ("y", "x"),
    driver: str = "GTiff",
) -> Union[xr.DataArray, xr.Dataset]:
    """Convert (Geo)Tiff byte responses to ``xarray.Dataset``.

    Parameters
    ----------
    r_dict : dict
        Dictionary of (Geo)Tiff byte responses where keys are some names that are used
        for naming each responses, and values are bytes.
    geometry : Polygon, MultiPolygon, or tuple
        The geometry to mask the data that should be in the same CRS as the r_dict.
    geo_crs : str
        The spatial reference of the input geometry.
    ds_dims : tuple of str, optional
        The names of the vertical and horizontal dimensions (in that order)
        of the target dataset, default to ("y", "x").
    driver : str, optional
        A GDAL driver for reading the content, defaults to GTiff. A list of the drivers
        can be found here: https://gdal.org/drivers/raster/index.html

    Returns
    -------
    xarray.Dataset or xarray.DataAraay
        Parallel (with dask) dataset or dataarray.
    """
    if not isinstance(r_dict, dict):
        raise InvalidInputType("r_dict", "dict", '{"name": Response.content}')  # noqa: FS003

    try:
        key1 = next(iter(r_dict.keys()))
    except StopIteration:
        raise StopIteration("Resnpose dict is empty.")

    if "_dd_" in key1:
        var_name = {lyr: "_".join(lyr.split("_")[:-3]) for lyr in r_dict.keys()}
    else:
        var_name = dict(zip(r_dict, r_dict))

    nodata, r_crs = _get_nodata_crs(r_dict[key1], driver)

    _geometry = _geo2polygon(geometry, geo_crs, r_crs)

    tmp_dir = tempfile.gettempdir()

    valid_ycoords = {"y", "Y", "lat", "Lat", "latitude", "Latitude"}
    valid_xcoords = {"x", "X", "lon", "Lon", "longitude", "Longitude"}

    def to_dataset(lyr: str, resp: bytes) -> xr.DataArray:
        with rio.MemoryFile() as memfile:
            memfile.write(resp)
            with memfile.open(driver=driver) as src:
                geom = [_geometry.intersection(sgeom.box(*src.bounds))]
                if geom[0].is_empty:
                    msk, transform, _ = rio_mask.raster_geometry_mask(src, [_geometry], invert=True)
                else:
                    msk, transform, _ = rio_mask.raster_geometry_mask(src, geom, invert=True)
                meta = src.meta
                meta.update(
                    {
                        "driver": driver,
                        "height": msk.shape[0],
                        "width": msk.shape[1],
                        "transform": transform,
                        "nodata": nodata,
                    }
                )

                with rio.vrt.WarpedVRT(src, **meta) as vrt:
                    ds = xr.open_rasterio(vrt)
                    valid_dims = list(ds.sizes)
                    if any(d not in valid_dims for d in ds_dims):
                        raise InvalidInputValue("ds_dims", valid_dims)
                    with contextlib.suppress(ValueError):
                        ds = ds.squeeze("band", drop=True)

                    coords = {
                        ds_dims[0]: ds.coords[ds_dims[0]],
                        ds_dims[1]: ds.coords[ds_dims[1]],
                    }
                    msk_da = xr.DataArray(msk, coords, dims=ds_dims)
                    ds = ds.where(msk_da, drop=True)
                    ycoord = list(set(ds.coords).intersection(valid_ycoords))[0]
                    ds = ds.sortby(ycoord[0], ascending=False)
                    ds.attrs["crs"] = r_crs.to_string()
                    ds.name = var_name[lyr]
                    fpath = Path(tmp_dir, f"{uuid.uuid4().hex}.nc")
                    ds.to_netcdf(fpath)
                    return fpath

    ds = xr.open_mfdataset((to_dataset(lyr, resp) for lyr, resp in r_dict.items()), parallel=True)

    if len(ds.variables) - len(ds.dims) == 1:
        ds = ds[list(ds.keys())[0]]

    ds.attrs["nodatavals"] = (nodata,)

    ycoord = list(set(ds.coords).intersection(valid_ycoords))
    xcoord = list(set(ds.coords).intersection(valid_xcoords))
    if len(xcoord) == 1 and len(ycoord) == 1:
        transform, _, _ = _get_transform(ds, ds_dims)
        ds = ds.sortby(ycoord[0], ascending=False)
        ds.attrs["transform"] = transform
        ds.attrs["res"] = (transform.a, transform.e)

    return ds


def _get_nodata_crs(resp: bytes, driver: str) -> Tuple[np.float64, pyproj.crs.crs.CRS]:
    """Get nodata and crs value of a raster in bytes.

    Parameters
    ----------
    resp : bytes
        Raster response returned from a wed service request such as WMS
    driver : str
        A GDAL driver for reading the content, defaults to GTiff. A list of the drivers
        can be found here: https://gdal.org/drivers/raster/index.html

    Returns
    -------
    tuple
        (NoData, CRS)
    """
    with rio.MemoryFile() as memfile:
        memfile.write(resp)
        with memfile.open(driver=driver) as src:
            r_crs = pyproj.CRS(src.crs)
            if src.nodata is None:
                try:
                    nodata = np.iinfo(src.dtypes[0]).max
                except ValueError:
                    nodata = np.nan
            else:
                nodata = np.dtype(src.dtypes[0]).type(src.nodata)
    return nodata, r_crs


def _get_transform(
    ds: Union[xr.Dataset, xr.DataArray],
    ds_dims: Tuple[str, str] = ("y", "x"),
) -> Tuple[affine.Affine, int, int]:
    """Get transform of a ``xarray.Dataset`` or ``xarray.DataArray``.

    Parameters
    ----------
    ds : xarray.Dataset or xarray.DataArray
        The dataset(array) to be masked
    ds_dims : tuple, optional
        Names of the coordinames in the dataset, defaults to ``("y", "x")``.

    Returns
    -------
    affine.Affine, int, int
        The affine transform, width, and height
    """
    ydim, xdim = ds_dims
    height, width = ds.sizes[ydim], ds.sizes[xdim]

    left, right = ds[xdim].min().item(), ds[xdim].max().item()
    bottom, top = ds[ydim].min().item(), ds[ydim].max().item()

    x_res = abs(left - right) / (width - 1)
    y_res = abs(top - bottom) / (height - 1)

    left -= x_res * 0.5
    right += x_res * 0.5
    top += y_res * 0.5
    bottom -= y_res * 0.5

    transform = rio_transform.from_bounds(left, bottom, right, top, width, height)
    return transform, width, height


def _geo2polygon(
    geometry: Union[Polygon, MultiPolygon, Tuple[float, float, float, float]],
    geo_crs: str,
    crs: str,
) -> Polygon:
    """Convert a geometry to a Shapely's Polygon and transform to any CRS.

    Parameters
    ----------
    geometry : Polygon or tuple of length 4
        Polygon or bounding box (west, south, east, north).
    geo_crs : str
        Spatial reference of the input geometry
    crs : str
        Target spatial reference.

    Returns
    -------
    Polygon
        A Polygon in the target CRS.
    """
    match_crs = MatchCRS(geo_crs, crs)
    geom = sgeom.box(*geometry) if isinstance(geometry, tuple) else geometry
    geom = match_crs.geometry(geom)

    if not geom.is_valid:
        geom = geom.buffer(0.0)
    return geom


class MatchCRS:
    """Reproject a geometry to another CRS.

    Parameters
    ----------
    in_crs : str
        Spatial reference of the input geometry
    out_crs : str
        Target spatial reference
    """

    def __init__(self, in_crs: str, out_crs: str):
        self.project = pyproj.Transformer.from_crs(in_crs, out_crs, always_xy=True).transform

    def geometry(
        self, geom: Union[Polygon, LineString, MultiLineString, MultiPolygon, Point, MultiPoint]
    ) -> Union[Polygon, LineString, MultiLineString, MultiPolygon, Point, MultiPoint]:
        """Reproject a geometry to the specified output CRS.

        Parameters
        ----------
        geom : LineString, MultiLineString, Polygon, MultiPolygon, Point, or MultiPoint
            Input geometry.

        Returns
        -------
        LineString, MultiLineString, Polygon, MultiPolygon, Point, or MultiPoint
            Input geometry in the specified CRS.

        Examples
        --------
        >>> from pygeoogc import MatchCRS
        >>> from shapely.geometry import Point
        >>> point = Point(-7766049.665, 5691929.739)
        >>> MatchCRS("epsg:3857", "epsg:4326").geometry(point).xy
        (array('d', [-69.7636111130079]), array('d', [45.44549114818127]))
        """
        if not isinstance(
            geom, (Polygon, LineString, MultiLineString, MultiPolygon, Point, MultiPoint)
        ):
            types = "LineString, MultiLineString, Polygon, MultiPolygon, Point, or MultiPoint"
            raise InvalidInputType("geom", types)

        return ops.transform(self.project, geom)

    def bounds(self, geom: Tuple[float, float, float, float]) -> Tuple[float, float, float, float]:
        """Reproject a bounding box to the specified output CRS.

        Parameters
        ----------
        geom : tuple
            Input bounding box (xmin, ymin, xmax, ymax).

        Returns
        -------
        tuple
            Input bounding box in the specified CRS.

        Examples
        --------
        >>> from pygeoogc import MatchCRS
        >>> bbox = (-7766049.665, 5691929.739, -7763049.665, 5696929.739)
        >>> MatchCRS("epsg:3857", "epsg:4326").bounds(bbox)
        (-69.7636111130079, 45.44549114818127, -69.73666165448431, 45.47699468552394)
        """
        if not (isinstance(geom, tuple) and len(geom) == 4):
            raise InvalidInputType("geom", "tuple", BOX_ORD)

        return ops.transform(self.project, sgeom.box(*geom)).bounds

    def coords(self, geom: List[Tuple[float, float]]) -> List[Tuple[Any, ...]]:
        """Reproject a list of coordinates to the specified output CRS.

        Parameters
        ----------
        geom : list of tuple
            Input coords [(x1, y1), ...].

        Returns
        -------
        tuple
            Input list of coords in the specified CRS.

        Examples
        --------
        >>> from pygeoogc import MatchCRS
        >>> coords = [(-7766049.665, 5691929.739)]
        >>> MatchCRS("epsg:3857", "epsg:4326").coords(coords)
        [(-69.7636111130079, 45.44549114818127)]
        """
        if not (isinstance(geom, list) and all(len(c) == 2 for c in geom)):
            raise InvalidInputType("geom", "list of tuples", "[(x1, y1), ...]")

        return list(zip(*self.project(*zip(*geom))))



"""Tests for PyGeoUtils"""
import io
import shutil

import rasterio
from pygeoogc import WFS, WMS
from shapely.geometry import Polygon

import pygeoutils as geoutils
from pygeoutils import MatchCRS

DEF_CRS = "epsg:4326"
ALT_CRS = "epsg:2149"
GEO_URB = Polygon(
    [
        [-118.72, 34.118],
        [-118.31, 34.118],
        [-118.31, 34.518],
        [-118.72, 34.518],
        [-118.72, 34.118],
    ]
)

GEO_NAT = Polygon(
    [[-69.77, 45.07], [-69.31, 45.07], [-69.31, 45.45], [-69.77, 45.45], [-69.77, 45.07]]
)


def test_json2geodf():
    url_wfs = "https://hazards.fema.gov/gis/nfhl/services/public/NFHL/MapServer/WFSServer"

    wfs = WFS(
        url_wfs,
        layer="public_NFHL:Base_Flood_Elevations",
        outformat="esrigeojson",
        crs="epsg:4269",
        version="2.0.0",
    )
    bbox = GEO_URB.bounds
    r = wfs.getfeature_bybox(bbox, box_crs=DEF_CRS)
    flood = geoutils.json2geodf([r.json(), r.json()], "epsg:4269", DEF_CRS)
    flood = flood.drop_duplicates()

    assert abs(flood["ELEV"].sum() - 781717.6) < 1e-1


def test_gtiff2array():
    url_wms = "https://www.mrlc.gov/geoserver/mrlc_download/wms"
    wms = WMS(
        url_wms,
        layers="NLCD_Land_Cover_Change_Index_Science_product_L48",
        outformat="image/geotiff",
        crs=DEF_CRS,
    )
    r_dict = wms.getmap_bybox(
        GEO_NAT.bounds,
        1e3,
        box_crs=DEF_CRS,
    )
    cover_box = geoutils.gtiff2xarray(r_dict, GEO_NAT.bounds, DEF_CRS)
    cover = geoutils.gtiff2xarray(r_dict, GEO_NAT, DEF_CRS)

    assert (
        abs(cover_box.mean().values.item() - 2.444) < 1e-3
        and abs(cover.mean().values.item() - 2.444) < 1e-3
    )


def test_envelope():
    resp = """
    {
      "attributes": {"OBJECTID": 5141},
      "geometry": {
      "xmin" : -109.55, "ymin" : 25.76, "xmax" : -86.39, "ymax" : 49.94,
      "zmin" : -12.0, "zmax" : 13.3,
      "spatialReference" : {"wkid" : 4326}}
    }"""

    expected = {
        "type": "Feature",
        "geometry": {
            "type": "Polygon",
            "coordinates": [
                [
                    [-86.39, 49.94],
                    [-109.55, 49.94],
                    [-109.55, 25.76],
                    [-86.39, 25.76],
                    [-86.39, 49.94],
                ]
            ],
        },
        "properties": {"OBJECTID": 5141},
        "id": 5141,
    }

    assert geoutils.arcgis2geojson(resp) == expected


def test_attr_none():
    resp = """
    {
      "geometry": {
      "xmin" : -109.55, "ymin" : 25.76, "xmax" : -86.39, "ymax" : 49.94,
      "zmin" : -12.0, "zmax" : 13.3,
      "spatialReference" : {"wkid" : 4326}}
    }"""
    expected = {
        "type": "Feature",
        "geometry": {
            "type": "Polygon",
            "coordinates": [
                [
                    [-86.39, 49.94],
                    [-109.55, 49.94],
                    [-109.55, 25.76],
                    [-86.39, 25.76],
                    [-86.39, 49.94],
                ]
            ],
        },
        "properties": None,
    }
    assert geoutils.arcgis2geojson(resp) == expected


def test_geometry_none():
    resp = '{"attributes": {"OBJECTID": 5141}}'
    expected = {"type": "Feature", "geometry": None, "properties": {"OBJECTID": 5141}, "id": 5141}
    assert geoutils.arcgis2geojson(resp) == expected


def test_no_id():
    resp = '{"attributes": {"OBJECTIDs": 5141}}'
    expected = {"type": "Feature", "geometry": None, "properties": {"OBJECTIDs": 5141}}
    assert geoutils.arcgis2geojson(resp) == expected


def test_ring():
    ring = {
        "rings": [
            [
                [-97.06138, 32.837],
                [-97.06133, 32.836],
                [-97.06124, 32.834],
                [-97.06127, 32.832],
                [-97.06138, 32.837],
            ],
            [[-97.06326, 32.759], [-97.06298, 32.755], [-97.06153, 32.749], [-97.06326, 32.759]],
        ],
        "spatialReference": {"wkid": 4326},
    }
    _ring = geoutils.arcgis2geojson(ring)
    res = {
        "type": "MultiPolygon",
        "coordinates": [
            [
                [
                    [-97.06138, 32.837],
                    [-97.06127, 32.832],
                    [-97.06124, 32.834],
                    [-97.06133, 32.836],
                    [-97.06138, 32.837],
                ]
            ],
            [[[-97.06326, 32.759], [-97.06298, 32.755], [-97.06153, 32.749], [-97.06326, 32.759]]],
        ],
    }
    assert _ring == res


def test_point():
    point = {"x": -118.15, "y": 33.80, "z": 10.0, "spatialReference": {"wkid": 4326}}
    _point = geoutils.arcgis2geojson(point)
    res = {"type": "Point", "coordinates": [-118.15, 33.8, 10.0]}
    assert _point == res


def test_multipoint():
    mpoint = {
        "hasZ": "true",
        "points": [
            [-97.06138, 32.837, 35.0],
            [-97.06133, 32.836, 35.1],
            [-97.06124, 32.834, 35.2],
        ],
        "spatialReference": {"wkid": 4326},
    }
    _mpoint = geoutils.arcgis2geojson(mpoint)
    res = {
        "type": "MultiPoint",
        "coordinates": [
            [-97.06138, 32.837, 35.0],
            [-97.06133, 32.836, 35.1],
            [-97.06124, 32.834, 35.2],
        ],
    }
    assert _mpoint == res


def test_path():
    path = {
        "hasM": "true",
        "paths": [
            [
                [-97.06138, 32.837, 5],
                [-97.06133, 32.836, 6],
                [-97.06124, 32.834, 7],
                [-97.06127, 32.832, 8],
            ],
            [[-97.06326, 32.759], [-97.06298, 32.755]],
        ],
        "spatialReference": {"wkid": 4326},
    }
    _path = geoutils.arcgis2geojson(path)
    res = {
        "type": "MultiLineString",
        "coordinates": [
            [
                [-97.06138, 32.837, 5],
                [-97.06133, 32.836, 6],
                [-97.06124, 32.834, 7],
                [-97.06127, 32.832, 8],
            ],
            [[-97.06326, 32.759], [-97.06298, 32.755]],
        ],
    }
    assert _path == res


def test_matchcrs():
    match_crs = MatchCRS(DEF_CRS, ALT_CRS)
    bounds = GEO_URB.bounds
    points = [(bounds[0], bounds[1]), (bounds[2], bounds[3])]
    coords = match_crs.coords(points)
    bbox = match_crs.bounds(GEO_URB.bounds)
    geom = match_crs.geometry(GEO_URB)
    assert (
        abs(geom.centroid.x * 1e-4 - (-362.099)) < 1e-3
        and abs(bbox[0] * 1e-4 - (-365.403)) < 1e-3
        and abs(coords[-1][0] * 1e-4 == (-287.707)) < 1e-3
    )


def test_show_versions():
    f = io.StringIO()
    geoutils.show_versions(file=f)
    assert "INSTALLED VERSIONS" in f.getvalue()
